"""
Tools to create audio
"""

from loopyng.gen import *
from loopyng.utils import *
from loopyng.sound import *
from loopyng.lite_sample_sounds import (
    tag_wf_gen,
    AnnotatedWaveform,
    chk_from_pattern,
    random_samples,
    pure_tone,
    triangular_tone,
    square_tone,
    tag_to_wf_gen_func,
    mk_audio_store,
)
